# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Easily declare tensorflow functions with custom gradients.

Normally, overriding gradients in TensorFlow requires you to register the
gradient function to a string and then use a with block with
graph.gradient_override_map().

There's a bunch of things about this that are annoying:

(1) You have to pick a string
(2) You can only use a string once, so if you're prototyping you need to
    generate a new name every time you modify the function.
(3) You have to make annoying with blocks all the time and it doesn't feel
    very functional.

This abstraction solves those issues. You no longer need to think about strings
or with blocks. Just use a single decorator and everything else will be
handled for you.

If you don't need to serialize your graph and the gradient override isn't
performance critical, you can use the high level `use_gradient()` decorator:

  @use_gradient(_foo_grad)
  def foo(x): ...

Otherwise, you can use use the lower level `gradient_override_map()`, a
convenience wrapper for `graph.gradient_override_map()`.
"""

from contextlib import contextmanager
import numpy as np
import tensorflow as tf
import uuid


def register_to_random_name(grad_f):
  """Register a gradient function to a random string.

  In order to use a custom gradient in TensorFlow, it must be registered to a
  string. This is both a hassle, and -- because only one function can every be
  registered to a string -- annoying to iterate on in an interactive
  environemnt.

  This function registers a function to a unique random string of the form:

    {FUNCTION_NAME}_{RANDOM_SALT}

  And then returns the random string. This is a helper in creating more
  convenient gradient overrides.

  Args:
    grad_f: gradient function to register. Should map (op, grad) -> grad(s)

  Returns:
    String that gradient function was registered to.
  """
  grad_f_name = grad_f.__name__ + "_" + str(uuid.uuid4())
  tf.RegisterGradient(grad_f_name)(grad_f)
  return grad_f_name


@contextmanager
def gradient_override_map(override_dict):
  """Convenience wrapper for graph.gradient_override_map().

  This functions provides two conveniences over normal tensorflow gradient
  overrides: it auomatically uses the default graph instead of you needing to
  find the graph, and it automatically

  Example:

    def _foo_grad_alt(op, grad): ...

    with gradient_override({"Foo": _foo_grad_alt}):

  Args:
    override_dict: A dictionary describing how to override the gradient.
      keys: strings correponding to the op type that should have their gradient
        overriden.
      values: functions or strings registered to gradient functions

  """
  override_dict_by_name = {}
  for (op_name, grad_f) in override_dict.items():
    if isinstance(grad_f, str):
       override_dict_by_name[op_name] = grad_f
    else:
      override_dict_by_name[op_name] = register_to_random_name(grad_f)
  with tf.get_default_graph().gradient_override_map(override_dict_by_name):
    yield


def use_gradient(grad_f):
  """Decorator for easily setting custom gradients for TensorFlow functions.

  * DO NOT use this function if you need to serialize your graph.
  * This function will cause the decorated function to run slower.

  Example:

    def _foo_grad(op, grad): ...

    @use_gradient(_foo_grad)
    def foo(x1, x2, x3): ...

  Args:
    grad_f: function to use as gradient.

  Returns:
    A decorator to apply to the function you wish to override the gradient of.

  """
  grad_f_name = register_to_random_name(grad_f)

  def function_wrapper(f):
    def inner(*inputs):

      # TensorFlow only supports (as of writing) overriding the gradient of
      # individual ops. In order to override the gardient of `f`, we need to
      # somehow make it appear to be an individual TensorFlow op.
      #
      # Our solution is to create a PyFunc that mimics `f`.
      #
      # In particular, we construct a graph for `f` and run it, then use a
      # stateful PyFunc to stash it's results in Python. Then we have another
      # PyFunc mimic it by taking all the same inputs and returning the stashed
      # output.
      #
      # I wish we could do this without PyFunc, but I don't see a way to have
      # it be fully general.

      state = {"out_value": None}

      # First, we need to run `f` and store it's output.

      out = f(*inputs)

      def store_out(out_value):
        """Store the value of out to a python variable."""
        state["out_value"] = out_value

      store_name = "store_" + f.__name__
      store = tf.py_func(store_out, [out], (), stateful=True, name=store_name)

      # Next, we create the mock function, with an overriden gradient.
      # Note that we need to make sure store gets evaluated before the mock
      # runs.

      def mock_f(*inputs):
        """Mimic f by retrieving the stored value of out."""
        return state["out_value"]

      with tf.control_dependencies([store]):
        with gradient_override_map({"PyFunc": grad_f_name}):
          mock_name = "mock_" + f.__name__
          mock_out = tf.py_func(mock_f, inputs, out.dtype, stateful=True,
                                name=mock_name)
          mock_out.set_shape(out.get_shape())

      # Finally, we can return the mock.

      return mock_out
    return inner
  return function_wrapper



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Methods for writing bytes to arbitrary destinations.

This module takes data and a URL, and attempts to save that data at that URL.

"""

from __future__ import absolute_import, division, print_function

from contextlib import contextmanager
from urllib.parse import urlparse
from tensorflow import gfile
import os


def _supports_make_dirs(path):
  """Whether this path implies a storage system that supports and requires
  intermediate directories to be created explicitly."""
  return not path.startswith("/bigstore")


def _supports_binary_writing(path):
  """Whether this path implies a storage system that supports and requires
  intermediate directories to be created explicitly."""
  return not path.startswith("/bigstore")


def _write_to_path(data, path, mode='wb'):
  with write_handle(path, mode) as handle:
    handle.write(data)


# Public functions


def write(data, url, mode='wb'):
  if urlparse(url).scheme in ('http', 'https'):
    message = "Writing to remote URL (%s) is not yet supported."
    raise ValueError(message, url)

  _write_to_path(data, url, mode=mode)


@contextmanager
def write_handle(path, mode=None):

  if _supports_make_dirs(path):
    gfile.MakeDirs(os.path.dirname(path))

  if mode is None:
    if _supports_binary_writing(path):
      mode = 'wb'
    else:
      mode = 'w'

  handle = gfile.Open(path, mode)
  yield handle
  handle.close()



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Objective functions for visualizing neural networks.

We represent objectives with a class `Objective` enclsoing functions of the
form:

  (T) => TensorFlow Scalar

Where `T` is a function that allows one to access the activations of different
layers of the network. For example `T("mixed4a")` gives the activations for
the layer mixed4a.

This allows objectives to be declared outside the rendering function, but then
actually constructed within its graph/session.
"""

from __future__ import absolute_import, division, print_function

from decorator import decorator
import numpy as np
import tensorflow as tf

# We use T as a variable name to access all kinds of tensors
# pylint: disable=invalid-name


class Objective(object):
  """"A wrapper to make objective functions easy to combine.

  For example, suppose you want to optimize 20% for mixed4a:20 and 80% for
  mixed4a:21. Then you could use:

    objetive = 0.2 * channel("mixed4a", 20) + 0.8 * channel("mixed4a", 0.8)

  Under the hood, we think of objectives as functions of the form:

    T => tensorflow scalar for loss

  where T is a function allowing you to index layers in the network -- that is,
  if there's a layer "mixed4a" then T("mixed4a") would give you its
  activations).

  This allows objectives to be declared outside the rendering function, but then
  actually constructed within its graph/session.
  """

  def __init__(self, objective_func, name="", description=""):
    self.objective_func = objective_func
    self.name = name
    self.description = description

  def __add__(self, other):
    if isinstance(other, (int, float)):
      objective_func = lambda T: other + self(T)
      name = self.name
      description = self.description
    else:
      objective_func = lambda T: self(T) + other(T)
      name = ", ".join([self.name, other.name])
      description = "Sum(" + " +\n".join([self.description, other.description]) + ")"
    return Objective(objective_func, name=name, description=description)

  def __neg__(self):
    return -1 * self

  def __sub__(self, other):
    return self + (-1 * other)

  @staticmethod
  def sum(objs):
    objective_func = lambda T: sum([obj(T) for obj in objs])
    descriptions = [obj.description for obj in objs]
    description = "Sum(" + " +\n".join(descriptions) + ")"
    names = [obj.name for obj in objs]
    name = ", ".join(names)
    return Objective(objective_func, name=name, description=description)

  def __mul__(self, other):
    if isinstance(other, (int, float)):
      objective_func = lambda T: other * self(T)
    else:
      objective_func = lambda T: self(T) * other(T)
    return Objective(objective_func, name=self.name, description=self.description)

  def __rmul__(self, other):
    return self.__mul__(other)

  def __radd__(self, other):
    return self.__add__(other)

  def __call__(self, T):
    return self.objective_func(T)


def _make_arg_str(arg):
  arg = str(arg)
  too_big = len(arg) > 15 or "\n" in arg
  return "..." if too_big else arg


@decorator
def wrap_objective(f, *args, **kwds):
  """Decorator for creating Objective factories.

  Changes f from the closure: (args) => () => TF Tensor
  into an Obejective factory: (args) => Objective

  while perserving function name, arg info, docs... for interactive python.
  """
  objective_func = f(*args, **kwds)
  objective_name = f.__name__
  args_str = " [" + ", ".join([_make_arg_str(arg) for arg in args]) + "]"
  description = objective_name.title() + args_str
  return Objective(objective_func, objective_name, description)


@wrap_objective
def neuron(layer_name, channel_n, x=None, y=None, batch=None):
  """Visualize a single neuron of a single channel."""
  def inner(T):
    layer = T(layer_name)
    shape = tf.shape(layer)
    if x is None:
      x_ = shape[1] // 2
    if y is None:
      y_ = shape[2] // 2
    if batch is None:
      return layer[:, x_, y_, channel_n]
    else:
      return layer[batch, x_, y_, channel_n]
  return inner


@wrap_objective
def channel(layer, n_channel, batch=None):
  """Visualize a single channel"""
  if batch is None:
    return lambda T: tf.reduce_mean(T(layer)[..., n_channel])
  else:
    return lambda T: tf.reduce_mean(T(layer)[batch, ..., n_channel])


@wrap_objective
def direction(layer, vec, batch=None):
  """Visualize a direction"""
  if batch is None:
    vec = vec[None, None, None]
    return lambda T: tf.reduce_mean(T(layer) * vec)
  else:
    vec = vec[None, None]
    return lambda T: tf.reduce_mean(T(layer)[batch] * vec)


@wrap_objective
def direction_neuron(layer_name, vec, batch=None, x=None, y=None):
  """Visualize a single (x, y) position along the given direction"""
  def inner(T):
    layer = T(layer_name)
    shape = tf.shape(layer)
    x_ = shape[1] / 2 if x is None else x
    y_ = shape[2] / 2 if y is None else y
    if batch is None:
      return tf.reduce_mean(layer[:, x_, y_] * vec[None])
    else:
      return tf.reduce_mean(layer[batch, x_, y_] * vec)
  return inner

@wrap_objective
def direction_cossim(layer, vec, batch=None):
  """Visualize a direction (cossine similarity)"""
  def inner(T):
    act_mags = tf.sqrt(tf.reduce_sum(T(layer)**2, -1, keep_dims=True))
    vec_mag = tf.sqrt(tf.reduce_sum(vec**2))
    mags = act_mags * vec_mag
    if batch is None:
      return tf.reduce_mean(T(layer) * vec.reshape([1, 1, 1, -1]) / mags)
    else:
      return tf.reduce_mean(T(layer)[batch] * vec.reshape([1, 1, -1]) / mags)
  return inner


@wrap_objective
def deepdream(layer):
  """Maximize 'interestingness' at some layer.

  See Mordvintsev et al., 2015.
  """
  return lambda T: tf.reduce_mean(T(layer)**2)


@wrap_objective
def total_variation(layer="input"):
  """Total variation of image (or activations at some layer).

  This operation is most often used as a penalty to reduce noise.
  See Simonyan, et al., 2014.
  """
  return lambda T: tf.image.total_variation(T(layer))


@wrap_objective
def L1(layer="input", constant=0, batch=None):
  """L1 norm of layer. Generally used as penalty."""
  if batch is None:
    return lambda T: tf.reduce_sum(tf.abs(T(layer) - constant))
  else:
    return lambda T: tf.reduce_sum(tf.abs(T(layer)[batch] - constant))


@wrap_objective
def L2(layer="input", constant=0, epsilon=1e-6, batch=None):
  """L2 norm of layer. Generally used as penalty."""
  if batch is None:
    return lambda T: tf.sqrt(epsilon + tf.reduce_sum((T(layer) - constant) ** 2))
  else:
    return lambda T: tf.sqrt(epsilon + tf.reduce_sum((T(layer)[batch] - constant) ** 2))


def _tf_blur(x, w=3):
  depth = x.shape[-1]
  k = np.zeros([w, w, depth, depth])
  for ch in range(depth):
    k_ch = k[:, :, ch, ch]
    k_ch[ :,    :  ] = 0.5
    k_ch[1:-1, 1:-1] = 1.0

  conv_k = lambda t: tf.nn.conv2d(t, k, [1, 1, 1, 1], "SAME")
  return conv_k(x) / conv_k(tf.ones_like(x))


@wrap_objective
def blur_input_each_step():
  """Minimizing this objective is equivelant to blurring input each step.

  Optimizing (-k)*blur_input_each_step() is equivelant to:

    input <- (1-k)*input + k*blur(input)

  An operation that was used in early feature visualization work.
  See Nguyen, et al., 2015.
  """
  def inner(T):
    t_input = T("input")
    t_input_blurred = tf.stop_gradient(_tf_blur(t_input))
    return 0.5*tf.reduce_sum((t_input - t_input_blurred)**2)
  return inner

@wrap_objective
def blur_alpha_each_step():
  """Minimizing this objective is equivelant to blurring input each step.

  Optimizing (-k)*blur_input_each_step() is equivelant to:

    input <- (1-k)*input + k*blur(input)

  An operation that was used in early feature visualization work.
  See Nguyen, et al., 2015.
  """
  def inner(T):
    t_input = T("input")[..., 3:4]
    t_input_blurred = tf.stop_gradient(_tf_blur(t_input))
    return 0.5*tf.reduce_sum((t_input - t_input_blurred)**2)
  return inner


@wrap_objective
def channel_interpolate(layer1, n_channel1, layer2, n_channel2):
  """Interpolate between layer1, n_channel1 and layer2, n_channel2.

  Optimize for a convex combination of layer1, n_channel1 and
  layer2, n_channel2, transitioning across the batch.

  Args:
    layer1: layer to optimize 100% at batch=0.
    n_channel1: neuron index to optimize 100% at batch=0.
    layer2: layer to optimize 100% at batch=N.
    n_channel2: neuron index to optimize 100% at batch=N.

  Returns:
    Objective
  """
  def inner(T):
    batch_n = T(layer1).get_shape().as_list()[0]
    arr1 = T(layer1)[..., n_channel1]
    arr2 = T(layer2)[..., n_channel2]
    weights = (np.arange(batch_n)/float(batch_n-1))
    S = 0
    for n in range(batch_n):
      S += (1-weights[n]) * tf.reduce_mean(arr1[n])
      S += weights[n] * tf.reduce_mean(arr2[n])
    return S
  return inner


@wrap_objective
def penalize_boundary_complexity(shp, w=20, mask=None, C=0.5):
  """Encourage the boundaries of an image to have less variation and of color C.

  Args:
    shp: shape of T("input") because this may not be known.
    w: width of boundary to penalize. Ignored if mask is set.
    mask: mask describing what area should be penalized.

  Returns:
    Objective.
  """
  def inner(T):
    arr = T("input")

    # print shp
    if mask is None:
      mask_ = np.ones(shp)
      mask_[:, w:-w, w:-w] = 0
    else:
      mask_ = mask

    blur = _tf_blur(arr, w=5)
    diffs = (blur-arr)**2
    diffs += 0.8*(arr-C)**2

    return -tf.reduce_sum(diffs*mask_)
  return inner


@wrap_objective
def alignment(layer, decay_ratio=2):
  """Encourage neighboring images to be similar.

  When visualizing the interpolation between two objectives, it's often
  desireable to encourage analagous boejcts to be drawn in the same position,
  to make them more comparable.

  This term penalizes L2 distance between neighboring images, as evaluated at
  layer.

  In general, we find this most effective if used with a paramaterization that
  shares across the batch. (In fact, that works quite well by iteself, so this
  function may just be obselete.)

  Args:
    layer: layer to penalize at.
    decay_ratio: how much to decay penalty as images move apart in batch.

  Returns:
    Objective.
  """
  def inner(T):
    batch_n = T(layer).get_shape().as_list()[0]
    arr = T(layer)
    accum = 0
    for d in [1, 2, 3, 4]:
      for i in range(batch_n - d):
        a, b = i, i+d
        arr1, arr2 = arr[a], arr[b]
        accum += tf.reduce_mean((arr1-arr2)**2) / decay_ratio**float(d)
    return -accum
  return inner

@wrap_objective
def diversity(layer):
  """Encourage diversity between each batch element.

  A neural net feature often responds to multiple things, but naive feature
  visualization often only shows us one. If you optimize a batch of images,
  this objective will encourage them all to be different.

  In particular, it caculuates the correlation matrix of activations at layer
  for each image, and then penalizes cossine similarity between them. This is
  very similar to ideas in style transfer, except we're *penalizing* style
  similarity instead of encouraging it.

  Args:
    layer: layer to evaluate activation correlations on.

  Returns:
    Objective.
  """
  def inner(T):
    layer_t = T(layer)
    batch_n, _, _, channels = layer_t.get_shape().as_list()

    flattened = tf.reshape(layer_t, [batch_n, -1, channels])
    grams = tf.matmul(flattened, flattened, transpose_a=True)
    grams = tf.nn.l2_normalize(grams, dim=[1,2], epsilon=1e-10)

    return sum([ sum([ tf.reduce_sum(grams[i]*grams[j])
                      for j in range(batch_n) if j != i])
                for i in range(batch_n)]) / batch_n
  return inner


@wrap_objective
def input_diff(orig_img):
  """Average L2 difference between optimized image and orig_img.

  This objective is usually mutliplied by a negative number and used as a
  penalty in making advarsarial counterexamples.
  """
  def inner(T):
    diff = T("input") - orig_img
    return tf.sqrt(tf.reduce_mean(diff**2))
  return inner


@wrap_objective
def class_logit(layer, label):
  """Like channel, but for softmax layers.

  Args:
    layer: A layer name string.
    label: Either a string (refering to a label in model.labels) or an int
      label position.

  Returns:
    Objective maximizing a logit.
  """
  def inner(T):
    if isinstance(label, int):
      class_n = label
    else:
      class_n = T("labels").index(label)
    logits = T(layer)
    logit = tf.reduce_sum(logits[:, class_n])
    return logit
  return inner


def as_objective(obj):
  """Convert obj into Objective class.

  Strings of the form "layer:n" become the Objective channel(layer, n).
  Objectives are returned unchanged.

  Args:
    obj: string or Objective.

  Returns:
    Objective
  """
  if isinstance(obj, Objective):
    return obj
  elif callable(obj):
    return obj
  elif isinstance(obj, str):
    layer, n = obj.split(":")
    layer, n = layer.strip(), int(n)
    return channel(layer, n)



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import tensorflow as tf
import numpy as np

from lucid.optvis.param.color import to_valid_rgb


def image_sample(shape, decorrelate=True, sd=None, decay_power=1):
  raw_spatial = rand_fft_image(shape, sd=sd, decay_power=decay_power)
  return to_valid_rgb(raw_spatial, decorrelate=decorrelate)

def rand_fft_image(shape, sd=None, decay_power=1):
  b, h, w, ch = shape
  sd = 0.01 if sd is None else sd
  
  imgs = []
  for _ in range(b):
    freqs = _rfft2d_freqs(h, w)
    fh, fw = freqs.shape
    spectrum_var = sd*tf.random_normal([2, ch, fh, fw], dtype="float32")
    spectrum = tf.complex(spectrum_var[0], spectrum_var[1])
    spertum_scale = 1.0 / np.maximum(freqs, 1.0/max(h, w))**decay_power
    # Scale the spectrum by the square-root of the number of pixels
    # to get a unitary transformation. This allows to use similar
    # leanring rates to pixel-wise optimisation.
    spertum_scale *= np.sqrt(w*h)
    scaled_spectrum = spectrum * spertum_scale
    img = tf.spectral.irfft2d(scaled_spectrum)
    # in case of odd input dimension we cut off the additional pixel
    # we get from irfft2d length computation
    img = img[:ch, :h, :w]
    img = tf.transpose(img, [1, 2, 0])
    imgs.append(img)
  return tf.stack(imgs)/4.

def _rfft2d_freqs(h, w):
  """Compute 2d spectrum frequences."""
  fy = np.fft.fftfreq(h)[:, None]
  # when we have an odd input dimension we need to keep one additional
  # frequency and later cut off 1 pixel
  if w % 2 == 1:
    fx = np.fft.fftfreq(w)[:w//2+2]
  else:
    fx = np.fft.fftfreq(w)[:w//2+1]
  return np.sqrt(fx*fx + fy*fy)



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Methods for read_handle bytes from arbitrary sources.

This module takes a URL, infers how to locate it,
loads the data into memory and returns it.
"""

from __future__ import absolute_import, division, print_function
from future.standard_library import install_aliases
install_aliases()

from contextlib import contextmanager
import os
import re
import logging
from urllib.parse import urlparse, urljoin
from future.moves.urllib import request
from tensorflow import gfile
from tempfile import gettempdir
from io import BytesIO, StringIO
import gc

from lucid.misc.io.writing import write, write_handle


# create logger with module name, e.g. lucid.misc.io.reading
log = logging.getLogger(__name__)


# Public functions


def read(url, encoding=None, cache=None):
  """Read from any URL.

  Internally differentiates between URLs supported by tf.gfile, such as URLs
  with the Google Cloud Storage scheme ('gs://...') or local paths, and HTTP
  URLs. This way users don't need to know about the underlying fetch mechanism.

  Args:
    url: a URL including scheme or a local path
    mode: mode in which to open the file. defaults to binary ('rb')
    encoding: if specified, encoding that should be used to decode read data
      if mode is specified to be text ('r'), this defaults to 'utf-8'.
    cache: whether to attempt caching the resource. Defaults to True only if
      the given URL specifies a remote resource.
  Returns:
    All bytes form the specified resource, or a decoded string of those.
  """
  with read_handle(url, cache) as handle:
    data = handle.read()

  if encoding:
    data = data.decode(encoding)

  return data


@contextmanager
def read_handle(url, cache=None):
  """Read from any URL with a file handle.

  Use this to get a handle to a file rather than eagerly load the data:

  ```
  with read_handle(url) as handle:
    result = something.load(handle)

  result.do_something()

  ```

  When program execution leaves this `with` block, the handle will be closed
  automatically.

  Args:
    url: a URL including scheme or a local path
  Returns:
    A file handle to the specified resource if it could be reached.
    The handle will be closed automatically once execution leaves this context.
  """
  scheme = urlparse(url).scheme

  if _is_remote(scheme) and cache is None:
    cache = True
    log.debug("Cache not specified, enabling because resource is remote.")

  if cache:
    handle = _read_and_cache(url)
  else:
    if scheme in ('http', 'https'):
      handle = _handle_web_url(url)
    elif scheme == 'gs':
      handle = _handle_gcs_url(url)
    else:
      handle = _handle_gfile(url)

  yield handle
  handle.close()


# Handlers


def _handle_gfile(url, mode='rb'):
  return gfile.Open(url, mode)


def _handle_web_url(url):
  return request.urlopen(url)


def _handle_gcs_url(url):
  # TODO: transparently allow authenticated access through storage API
  _, resource_name = url.split('://')
  base_url = 'https://storage.googleapis.com/'
  url = urljoin(base_url, resource_name)
  return _handle_web_url(url)


# Helper Functions


def _is_remote(scheme):
  return scheme in ('http', 'https', 'gs')


RESERVED_PATH_CHARS = re.compile("[^a-zA-Z0-9]")


def _read_and_cache(url):
  local_name = RESERVED_PATH_CHARS.sub('_', url)
  local_path = os.path.join(gettempdir(), local_name)
  if os.path.exists(local_path):
    log.info("Found cached file '%s'.", local_path)
    return _handle_gfile(local_path)
  else:
    log.info("Caching URL '%s' locally at '%s'.", url, local_path)
    with write_handle(local_path, 'wb') as output, read_handle(url, cache=False) as input:
      for chunk in _file_chunk_iterator(input):
        output.write(chunk)
    gc.collect()
    return _handle_gfile(local_path)


from functools import partial
from io import DEFAULT_BUFFER_SIZE
import sys

def _file_chunk_iterator(file_handle):
  reader = partial(file_handle.read1, DEFAULT_BUFFER_SIZE)
  file_iterator = iter(reader, bytes())
  # TODO: once dropping Python <3.3 compat, update to `yield from ...`
  for chunk in file_iterator:
    yield chunk



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from __future__ import absolute_import, division, print_function
from os import path

import tensorflow as tf
from lucid.modelzoo.util import load_text_labels, load_graphdef, forget_xy
from lucid.misc.io import load

class Model(object):
  """Base pretrained model importer."""

  model_path = None
  labels_path = None
  labels = None
  image_value_range = (-1, 1)
  image_shape = [None, None, 3]

  def __init__(self):
    self.graph_def = None
    if self.labels_path is not None:
      self.labels = load_text_labels(self.labels_path)

  def load_graphdef(self):
    self.graph_def = load_graphdef(self.model_path)

  def post_import(self, scope):
    pass

  def create_input(self, t_input=None, forget_xy_shape=True):
    """Create input tensor."""
    if t_input is None:
      t_input = tf.placeholder(tf.float32, self.image_shape)
    t_prep_input = t_input
    if len(t_prep_input.shape) == 3:
      t_prep_input = tf.expand_dims(t_prep_input, 0)
    if forget_xy_shape:
      t_prep_input = forget_xy(t_prep_input)
    lo, hi = self.image_value_range
    t_prep_input = lo + t_prep_input * (hi-lo)
    return t_input, t_prep_input

  def import_graph(self, t_input=None, scope='import', forget_xy_shape=True):
    """Import model GraphDef into the current graph."""
    graph = tf.get_default_graph()
    assert graph.unique_name(scope, False) == scope, (
        'Scope "%s" already exists. Provide explicit scope names when '
        'importing multiple instances of the model.') % scope
    t_input, t_prep_input = self.create_input(t_input, forget_xy_shape)
    tf.import_graph_def(
        self.graph_def, {self.input_name: t_prep_input}, name=scope)
    self.post_import(scope)


class SerializedModel(Model):
  """Allows importing various types of serialized models from a directory.

  (Currently only supports frozen graph models and relies on manifest.json file.
  In the future we may want to support automatically detecting the type and
  support loading more ways of saving models: tf.SavedModel, metagraphs, etc.)
  """

  @classmethod
  def from_directory(cls, model_path):
    manifest_path = path.join(model_path, 'manifest.json')
    try:
      manifest = load(manifest_path)
    except Exception as e:
      raise ValueError("Could not find manifest.json file in dir {}. Error: {}".format(model_path, e))
    if manifest.get('type', 'frozen') == 'frozen':
      return FrozenGraphModel(model_path, manifest)
    else: # TODO: add tf.SavedModel support, etc
      raise NotImplementedError("SerializedModel Manifest type '{}' has not been implemented!".format(manifest.get('type')))


class FrozenGraphModel(SerializedModel):

  def __init__(self, model_directory, manifest):
    model_path = manifest.get('model_path', 'graph.pb')
    if model_path.startswith("./"): # TODO: can we be less specific here?
      self.model_path = path.join(model_directory, model_path)
    else:
      self.model_path = model_path
    self.labels_path = manifest.get('labels_path', None)
    self.image_value_range = manifest.get('image_value_range', None)
    self.image_shape = manifest.get('image_shape', None)
    self.input_name = manifest.get('input_name', 'input:0')
    super().__init__()




# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""High-level wrapper for paramaterizing images."""


import numpy as np
import tensorflow as tf

from lucid.optvis.param.color import to_valid_rgb
from lucid.optvis.param.spatial import naive, fft_image

def image(w, h=None, batch=None, sd=None, decorrelate=True, fft=True, alpha=False):
  h = h or w
  batch = batch or 1
  channels = 4 if alpha else 3
  shape = [batch, w, h, channels]
  param_f = fft_image if fft else naive
  t = param_f(shape, sd=sd)
  rgb = to_valid_rgb(t[..., :3], decorrelate=decorrelate, sigmoid=True)
  if alpha:
    a = tf.nn.sigmoid(t[..., 3:])
    return tf.concat([rgb, a], -1)
  return rgb



"""3D mesh manipulation utilities."""

from builtins import str
from collections import OrderedDict
import numpy as np


def frustum(left, right, bottom, top, znear, zfar):
  """Create view frustum matrix."""
  assert right != left
  assert bottom != top
  assert znear != zfar

  M = np.zeros((4, 4), dtype=np.float32)
  M[0, 0] = +2.0 * znear / (right - left)
  M[2, 0] = (right + left) / (right - left)
  M[1, 1] = +2.0 * znear / (top - bottom)
  M[3, 1] = (top + bottom) / (top - bottom)
  M[2, 2] = -(zfar + znear) / (zfar - znear)
  M[3, 2] = -2.0 * znear * zfar / (zfar - znear)
  M[2, 3] = -1.0
  return M


def perspective(fovy, aspect, znear, zfar):
  """Create perspective projection matrix."""
  assert znear != zfar
  h = np.tan(fovy / 360.0 * np.pi) * znear
  w = h * aspect
  return frustum(-w, w, -h, h, znear, zfar)


def anorm(x, axis=None, keepdims=False):
  """Compute L2 norms alogn specified axes."""
  return np.sqrt((x*x).sum(axis=axis, keepdims=keepdims))


def normalize(v, axis=None, eps=1e-10):
  """L2 Normalize along specified axes."""
  return v / max(anorm(v, axis=axis, keepdims=True), eps)


def lookat(eye, target=[0, 0, 0], up=[0, 1, 0]):
  """Generate LookAt modelview matrix."""
  eye = np.float32(eye)
  forward = normalize(target - eye)
  side = normalize(np.cross(forward, up))
  up = np.cross(side, forward)
  M = np.eye(4, dtype=np.float32)
  R = M[:3, :3]
  R[:] = [side, up, -forward]
  M[:3, 3] = -R.dot(eye)
  return M


def homotrans(M, p):
  p = np.asarray(p)
  if p.shape[-1] == M.shape[1]-1:
    p = np.append(p, np.ones_like(p[...,:1]), -1)
  p = np.dot(p, M.T)
  return p[...,:-1] / p[...,-1:]


def _parse_vertex_tuple(s):
  """Parse vertex indices in '/' separated form (like 'i/j/k', 'i//k' ...)."""
  vt = [0, 0, 0]
  for i, c in enumerate(s.split('/')):
    if c:
      vt[i] = int(c)
  return tuple(vt)


def _unify_rows(a):
  """Unify lengths of each row of a."""
  lens = np.fromiter(map(len, a), np.int32)
  if not (lens[0] == lens).all():
    out = np.zeros((len(a), lens.max()), np.float32)
    for i, row in enumerate(a):
      out[i, :lens[i]] = row
  else:
    out = np.float32(a)
  return out


def load_obj(fn):
  """Load 3d mesh form .obj' file.
  
  Args:
    fn: Input file name or file-like object.
    
  Returns:
    dictionary with the following keys (some of which may be missing):
      position: np.float32, (n, 3) array, vertex positions
      uv: np.float32, (n, 2) array, vertex uv coordinates
      normal: np.float32, (n, 3) array, vertex uv normals
      face: np.int32, (k*3,) traingular face indices
  """
  position = [np.zeros(3, dtype=np.float32)]
  normal = [np.zeros(3, dtype=np.float32)]
  uv = [np.zeros(2, dtype=np.float32)]
  
  tuple2idx = OrderedDict()
  trinagle_indices = []
  
  input_file = open(fn) if isinstance(fn, str) else fn
  for line in input_file:
    line = line.strip()
    if not line or line[0] == '#':
      continue
    tag, line = line.split(' ', 1)
    if tag == 'v':
      position.append(np.fromstring(line, sep=' '))
    elif tag == 'vt':
      uv.append(np.fromstring(line, sep=' '))
    elif tag == 'vn':
      normal.append(np.fromstring(line, sep=' '))
    elif tag == 'f':
      output_face_indices = []
      for chunk in line.split():
        # tuple order: pos_idx, uv_idx, normal_idx
        vt = _parse_vertex_tuple(chunk)
        if vt not in tuple2idx:  # create a new output vertex?
          tuple2idx[vt] = len(tuple2idx)
        output_face_indices.append(tuple2idx[vt])
      # generate face triangles
      for i in range(1, len(output_face_indices)-1):
        for vi in [0, i, i+1]:
          trinagle_indices.append(output_face_indices[vi])
  
  outputs = {}
  outputs['face'] = np.int32(trinagle_indices)
  pos_idx, uv_idx, normal_idx = np.int32(list(tuple2idx)).T
  if np.any(pos_idx):
    outputs['position'] = _unify_rows(position)[pos_idx]
  if np.any(uv_idx):
    outputs['uv'] = _unify_rows(uv)[uv_idx]
  if np.any(normal_idx):
    outputs['normal'] = _unify_rows(normal)[normal_idx]
  return outputs



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Ensuring compatibilty across environments, e.g. Jupyter/Colab/Shell."""

from __future__ import absolute_import, division, print_function


def is_notebook_environment():
  try:
    shell = get_ipython().__class__.__name__
    if shell == 'ZMQInteractiveShell':
      return True   # IPython Notebook
    elif shell == 'Shell':
      return True   # Colaboratory Notebook
    elif shell == 'TerminalInteractiveShell':
      return False  # Terminal running IPython
    else:
      return False  # Other unknown type (?)
  except NameError:
    return False    # Probably standard Python interpreter



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Helper for using sklearn.decomposition on high-dimensional tensors.

Provides ChannelReducer, a wrapper around sklearn.decomposition to help them
apply to arbitrary rank tensors. It saves lots of annoying reshaping.
"""

import numpy as np
import sklearn.decomposition

class ChannelReducer(object):
  """Helper for dimensionality reduction to the innermost dimension of a tensor.

  This class wraps sklearn.decomposition classes to help them apply to arbitrary
  rank tensors. It saves lots of annoying reshaping.

  See the original sklearn.decomposition documentation:
  http://scikit-learn.org/stable/modules/classes.html#module-sklearn.decomposition
  """

  def __init__(self, n_features=3, reduction_alg="NMF", **kwargs):
    """Constructor for ChannelReducer.

    Inputs:
      n_features: Numer of dimensions to reduce inner most dimension to.
      reduction_alg: A string or sklearn.decomposition class. Defaults to
        "NMF" (non-negative matrix facotrization). Other options include:
        "PCA", "FastICA", and "MiniBatchDictionaryLearning". The name of any of
        the sklearn.decomposition classes will work, though.
      kwargs: Additional kwargs to be passed on to the reducer.
    """
    if isinstance(reduction_alg, str):
      reduction_alg = sklearn.decomposition.__getattribute__(reduction_alg)
    self.n_features = n_features
    self._reducer = reduction_alg(n_features, **kwargs)

  @classmethod
  def _apply_flat(cls, f, acts):
    """Utility for applying f to inner dimension of acts.

    Flattens acts into a 2D tensor, applies f, then unflattens so that all
    dimesnions except innermost are unchanged.
    """
    orig_shape = acts.shape
    acts_flat = acts.reshape([-1, acts.shape[-1]])
    new_flat = f(acts_flat)
    if not isinstance(new_flat, np.ndarray):
      return new_flat
    shape = list(orig_shape[:-1]) + [-1]
    return new_flat.reshape(shape)

  def fit(self, acts):
    return ChannelReducer._apply_flat(self._reducer.fit, acts)

  def fit_transform(self, acts):
    return ChannelReducer._apply_flat(self._reducer.fit_transform, acts)

  def transform(self, acts):
    return ChannelReducer._apply_flat(self._reducer.transform, acts)

  #TODO: decide whether to expose _reducer params higher up, figure
  # out any transforms that need to be done.

  #def __getattr__(self, name):
  #  return getattr(self._reducer, name)

  def __dir__(self):
    dynamic_attrs = dir(self._reducer)
    return self.__dict__.keys()# + dynamic_attrs



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf

from lucid.optvis.param.lowres import lowres_tensor


def naive(shape, sd=None):
  return lowres_tensor(shape, shape, sd=sd)


def _rfft2d_freqs(h, w):
  """Compute 2d spectrum frequences."""
  fy = np.fft.fftfreq(h)[:, None]
  # when we have an odd input dimension we need to keep one additional
  # frequency and later cut off 1 pixel
  if w % 2 == 1:
    fx = np.fft.fftfreq(w)[:w//2+2]
  else:
    fx = np.fft.fftfreq(w)[:w//2+1]
  return np.sqrt(fx*fx + fy*fy)


def fft_image(shape, sd=None, decay_power=1):
  b, h, w, ch = shape
  imgs = []
  for _ in range(b):
    freqs = _rfft2d_freqs(h, w)
    fh, fw = freqs.shape
    sd = sd or 0.01
    init_val = sd*np.random.randn(2, ch, fh, fw).astype("float32")
    spectrum_var = tf.Variable(init_val)
    spectrum = tf.complex(spectrum_var[0], spectrum_var[1])
    spertum_scale = 1.0 / np.maximum(freqs, 1.0/max(h, w))**decay_power
    # Scale the spectrum by the square-root of the number of pixels
    # to get a unitary transformation. This allows to use similar
    # leanring rates to pixel-wise optimisation.
    spertum_scale *= np.sqrt(w*h)
    scaled_spectrum = spectrum * spertum_scale
    img = tf.spectral.irfft2d(scaled_spectrum)
    # in case of odd input dimension we cut off the additional pixel
    # we get from irfft2d length computation
    img = img[:ch, :h, :w]
    img = tf.transpose(img, [1, 2, 0])
    imgs.append(img)
  return tf.stack(imgs)/4.


def laplacian_pyramid(shape, n_levels=4, sd=None):
  """Simple laplacian pyramid paramaterization of an image.

  For more flexibility, use a sum of lowres_tensor()s.

  Args:
    shape: shape of resulting image, [batch, width, height, channels].
    n_levels: number of levels of laplacian pyarmid.
    sd: standard deviation of param initialization.

  Returns:
    tensor with shape from first argument.
  """
  batch_dims = shape[:-3]
  w, h, ch = shape[-3:]
  pyramid = 0
  for n in range(n_levels):
    k = 2**n
    pyramid += lowres_tensor(shape, batch_dims + [w // k, h // k, ch], sd=sd)
  return pyramid


def sample_bilinear(texture, uv):
  """Build bilinear texture sampling graph.

  Coordinate transformation rules match OpenGL GL_REPEAT wrapping and GL_LINEAR
  interpolation modes.

  Args:
    texture: [tex_h, tex_w, channel_n] tensor.
    uv: [frame_h, frame_h, 2] tensor with per-pixel UV coordinates in range [0..1]

  Returns:
    [frame_h, frame_h, channel_n] tensor with per-pixel sampled values.
  """
  h, w = tf.unstack(tf.shape(texture)[:2])
  u, v = tf.split(uv, 2, axis=-1)
  v = 1.0-v  # vertical flip to match GL convention
  u, v = u*tf.to_float(w)-0.5, v*tf.to_float(h)-0.5
  u0, u1 = tf.floor(u), tf.ceil(u)
  v0, v1 = tf.floor(v), tf.ceil(v)
  uf, vf = u-u0, v-v0
  u0, u1, v0, v1 = map(tf.to_int32, [u0, u1, v0, v1])
  def sample(u, v):
    vu = tf.concat([v%h, u%w], axis=-1)
    return tf.gather_nd(texture, vu)
  s00, s01 = sample(u0, v0), sample(u0, v1)
  s10, s11 = sample(u1, v0), sample(u1, v1)
  s0 = s00*(1.0-vf) + s01*vf
  s1 = s10*(1.0-vf) + s11*vf
  s = s0*(1.0-uf) + s1*uf
  return s



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Methods for loading arbitrary data from arbitrary sources.

This module takes a URL, infers its underlying data type and how to locate it,
loads the data into memory and returns a convenient representation.

This should support for example PNG images, JSON files, npy files, etc.
"""

from __future__ import absolute_import, division, print_function

import os
import json
import logging
import numpy as np
import PIL.Image

from lucid.misc.io.reading import read_handle


# create logger with module name, e.g. lucid.misc.io.reading
log = logging.getLogger(__name__)


def _load_npy(handle, **kwargs):
  """Load npy file as numpy array."""
  del kwargs
  return np.load(handle)


def _load_img(handle, **kwargs):
  """Load image file as numpy array."""
  del kwargs
  # PIL.Image will infer image type from provided handle's file extension
  pil_img = PIL.Image.open(handle)
  img_array = np.asarray(pil_img)
  dtype = img_array.dtype
  # using np.divide should avoid an extra copy compared to doing division first
  return np.divide(img_array, np.iinfo(dtype).max, dtype=np.float32)


def _load_json(handle, **kwargs):
  """Load json file as python object."""
  del kwargs
  return json.load(handle)


def _load_text(handle, encoding='utf-8'):
  """Load and decode a string."""
  return handle.read().decode(encoding)

loaders = {
  ".png":  _load_img,
  ".jpg":  _load_img,
  ".jpeg": _load_img,
  ".npy":  _load_npy,
  ".npz":  _load_npy,
  ".json": _load_json,
  ".txt": _load_text,
  ".md": _load_text,
}


def load(url_or_handle, cache=None, encoding='utf-8'):
  """Load a file.

  File format is inferred from url. File retrieval strategy is inferred from
  URL. Returned object type is inferred from url extension.

  Args:
    url_or_handle: a (reachable) URL, or an already open file handle

  Raises:
    RuntimeError: If file extension or URL is not supported.
  """
  is_handle = hasattr(url_or_handle, 'read') and hasattr(url_or_handle, 'name')
  if is_handle:
    _, ext = os.path.splitext(url_or_handle.name)
  else:
    _, ext = os.path.splitext(url_or_handle)
  if not ext:
    raise RuntimeError("No extension in URL: " + url_or_handle)

  ext = ext.lower()
  if ext in loaders:
    loader = loaders[ext]
    message = "Using inferred loader '%s' due to passed file extension '%s'."
    log.debug(message, loader.__name__[6:], ext)
    if is_handle:
      result = loader(url_or_handle, encoding=encoding)
    else:
      with read_handle(url_or_handle, cache=cache) as handle:
        result = loader(handle, encoding=encoding)
    return result
  else:
    # TODO: handle 'handle' case? refactor?
    log.warn("Unknown extension '%s', attempting to load as image.", ext)
    try:
      with read_handle(url_or_handle, cache=cache) as handle:
        result = _load_img(handle)
    except Exception as e:
      message = "Could not load resource %s as image. Supported extensions: %s"
      log.error(message, url_or_handle, list(loaders))
      raise RuntimeError(message.format(url_or_handle, list(loaders)))
    else:
      log.info("Unknown extension '%s' successfully loaded as image.", ext)
      return result



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities for normalizing arrays and converting them to images."""

from __future__ import absolute_import, division, print_function

import base64
import logging
import numpy as np
import PIL.Image
from io import BytesIO


# create logger with module name, e.g. lucid.misc.io.array_to_image
log = logging.getLogger(__name__)


def _normalize_array(array, domain=(0, 1)):
  """Given an arbitrary rank-3 NumPy array, produce one representing an image.

  This ensures the resulting array has a dtype of uint8 and a domain of 0-255.

  Args:
    array: NumPy array representing the image
    domain: expected range of values in array,
      defaults to (0, 1), if explicitly set to None will use the array's
      own range of values and normalize them.

  Returns:
    normalized PIL.Image
  """
  # first copy the input so we're never mutating the user's data
  array = np.array(array)
  # squeeze helps both with batch=1 and B/W and PIL's mode inference
  array = np.squeeze(array)
  assert len(array.shape) <= 3
  assert np.issubdtype(array.dtype, np.number)

  low, high = np.min(array), np.max(array)
  if domain is None:
    message = "No domain specified, normalizing from measured (~%.2f, ~%.2f)"
    log.debug(message, low, high)
    domain = (low, high)

  # clip values if domain was specified and array contains values outside of it
  if low < domain[0] or high > domain[1]:
    message = "Clipping domain from (~{:.2f}, ~{:.2f}) to (~{:.2f}, ~{:.2f})."
    log.info(message.format(low, high, domain[0], domain[1]))
    array = array.clip(*domain)

  min_value, max_value = np.iinfo(np.uint8).min, np.iinfo(np.uint8).max  # 0, 255
  # convert signed to unsigned if needed
  if np.issubdtype(array.dtype, np.inexact):
    offset = domain[0]
    if offset != 0:
      array -= offset
      log.debug("Converting inexact array by subtracting -%.2f.", offset)
    scalar = max_value / (domain[1] - domain[0])
    if scalar != 1:
      array *= scalar
      log.debug("Converting inexact array by scaling by %.2f.", scalar)

  return array.clip(min_value, max_value).astype(np.uint8)


def _serialize_normalized_array(array, fmt='png', quality=70):
  """Given a normalized array, returns byte representation of image encoding.

  Args:
    array: NumPy array of dtype uint8 and range 0 to 255
    fmt: string describing desired file format, defaults to 'png'
    quality: specifies compression quality from 0 to 100 for lossy formats

  Returns:
    image data as BytesIO buffer
  """
  dtype = array.dtype
  assert np.issubdtype(dtype, np.unsignedinteger)
  assert np.max(array) <= np.iinfo(dtype).max
  assert array.shape[-1] > 1  # array dims must have been squeezed

  image = PIL.Image.fromarray(array)
  image_bytes = BytesIO()
  image.save(image_bytes, fmt, quality=quality)
  # TODO: Python 3 could save a copy here by using `getbuffer()` instead.
  image_data = image_bytes.getvalue()
  return image_data


def serialize_array(array, domain=(0, 1), fmt='png', quality=70):
  """Given an arbitrary rank-3 NumPy array,
  returns the byte representation of the encoded image.

  Args:
    array: NumPy array of dtype uint8 and range 0 to 255
    domain: expected range of values in array, see `_normalize_array()`
    fmt: string describing desired file format, defaults to 'png'
    quality: specifies compression quality from 0 to 100 for lossy formats

  Returns:
    image data as BytesIO buffer
  """
  normalized = _normalize_array(array, domain=domain)
  return _serialize_normalized_array(normalized, fmt=fmt, quality=quality)


JS_ARRAY_TYPES = {
    'int8', 'int16', 'int32', 'uint8', 'uint16', 'uint32', 'float32', 'float64'
}


def array_to_jsbuffer(array):
  """Serialize 1d NumPy array to JS TypedArray.

  Data is serialized to base64-encoded string, which is much faster
  and memory-efficient than json list serialization.

  Args:
    array: 1d NumPy array, dtype must be one of JS_ARRAY_TYPES.

  Returns:
    JS code that evaluates to a TypedArray as string.

  Raises:
    TypeError: if array dtype or shape not supported.
  """
  if array.ndim != 1:
    raise TypeError('Only 1d arrays can be converted JS TypedArray.')
  if array.dtype.name not in JS_ARRAY_TYPES:
    raise TypeError('Array dtype not supported by JS TypedArray.')
  js_type_name = array.dtype.name.capitalize() + 'Array'
  data_base64 = base64.b64encode(array.tobytes()).decode('ascii')
  code = """
    (function() {
      const data = atob("%s");
      const buf = new Uint8Array(data.length);
      for (var i=0; i<data.length; ++i) {
        buf[i] = data.charCodeAt(i);
      }
      var array_type = %s;
      if (array_type == Uint8Array) {
        return buf;
      }
      return new array_type(buf.buffer);
    })()
  """ % (data_base64, js_type_name)
  return code



import random
import json
import tempfile
import subprocess
import os.path as osp
import uuid

from IPython.core.magic import register_cell_magic

from lucid.misc.io.showing import _display_html
from lucid.misc.io.reading import read

_svelte_temp_dir = tempfile.mkdtemp(prefix="svelte_")

_template = """
  <div id='$id'></div>
  <script>
  $js
  </script>
  <script>
    var app = new $name({
        target: document.querySelector('#$id'),
        data: $data,
      });
  </script>
 """


def build_svelte(html_fname):
  js_fname = html_fname.replace(".html", ".js")
  cmd = "svelte compile --format iife " + html_fname + " > " + js_fname
  print(cmd)
  try:
    print(subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT))
  except subprocess.CalledProcessError as exception:
    print("Svelte build failed! Output:\n{}".format(exception.output.decode()))
  return js_fname


def SvelteComponent(name, path):
  """Display svelte components in iPython.

  Args:
    name: name of svelte component (must match component filename when built)
    path: path to compile svelte .js file or source svelte .html file.
      (If html file, we try to call svelte and build the file.)

  Returns:
    A function mapping data to a rendered svelte component in ipython.
  """
  if path[-3:] == ".js":
    js_path = path
  elif path[-5:] == ".html":
    print("Trying to build svelte component from html...")
    js_path = build_svelte(path)
  js_content = read(js_path)
  def inner(data):
    id_str = name + "_" + str(uuid.uuid4())
    html = _template \
        .replace("$js", js_content) \
        .replace("$name", name) \
        .replace("$data", json.dumps(data)) \
        .replace("$id", id_str)
    _display_html(html)
  return inner


@register_cell_magic
def html_define_svelte(line, cell):
  base_name = line.split()[0]
  name_str = base_name + "_" + str(uuid.uuid4())
  html_fname = osp.join(_svelte_temp_dir, name_str + ".html")
  with open(html_fname, "w") as f:
    f.write(cell)
  component = SvelteComponent(name_str, html_fname)
  globals()[base_name] = component



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import numpy as np
import tensorflow as tf

from lucid.optvis.param import lowres_tensor


def multi_interpolation_basis(n_objectives=6, n_interp_steps=5, width=128,
                              channels=3):
  """A paramaterization for interpolating between each pair of N objectives.

  Sometimes you want to interpolate between optimizing a bunch of objectives,
  in a paramaterization that encourages images to align.

  Args:
    n_objectives: number of objectives you want interpolate between
    n_interp_steps: number of interpolation steps
    width: width of intepolated images
    channel

  Returns:
    A [n_objectives, n_objectives, n_interp_steps, width, width, channel]
    shaped tensor, t, where the final [width, width, channel] should be
    seen as images, such that the following properties hold:

     t[a, b]    = t[b, a, ::-1]
     t[a, i, 0] = t[a, j, 0] for all i, j
     t[a, a, i] = t[a, a, j] for all i, j
     t[a, b, i] = t[b, a, -i] for all i

  """
  N, M, W, Ch = n_objectives, n_interp_steps, width, channels

  const_term = sum([lowres_tensor([W, W, Ch], [W//k, W//k, Ch])
                    for k in [1, 2, 4, 8]])
  const_term = tf.reshape(const_term, [1, 1, 1, W, W, Ch])

  example_interps = [
      sum([lowres_tensor([M, W, W, Ch], [2, W//k, W//k, Ch])
           for k in [1, 2, 4, 8]])
      for _ in range(N)]

  example_basis = []
  for n in range(N):
    col = []
    for m in range(N):
      interp = example_interps[n] + example_interps[m][::-1]
      col.append(interp)
    example_basis.append(col)

  interp_basis = []
  for n in range(N):
    col = [interp_basis[m][N-n][::-1] for m in range(n)]
    col.append(tf.zeros([M, W, W, 3]))
    for m in range(n+1, N):
      interp = sum([lowres_tensor([M, W, W, Ch], [M, W//k, W//k, Ch])
                    for k in [1, 2]])
      col.append(interp)
    interp_basis.append(col)

  basis = []
  for n in range(N):
    col_ex = tf.stack(example_basis[n])
    col_in = tf.stack(interp_basis[n])
    basis.append(col_ex + col_in)
  basis = tf.stack(basis)

  return basis + const_term



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Methods for displaying images from Numpy arrays."""

from __future__ import absolute_import, division, print_function

from io import BytesIO
import base64
import logging
import numpy as np
import IPython.display
from string import Template

from lucid.misc.io.serialize_array import serialize_array, array_to_jsbuffer


# create logger with module name, e.g. lucid.misc.io.showing
log = logging.getLogger(__name__)


def _display_html(html_str):
  IPython.display.display(IPython.display.HTML(html_str))


def _image_url(array, fmt='png', mode="data", quality=70, domain=None):
  """Create a data URL representing an image from a PIL.Image.

  Args:
    image: a numpy
    mode: presently only supports "data" for data URL

  Returns:
    URL representing image
  """
  # TODO: think about supporting saving to CNS, potential params: cns, name
  # TODO: think about saving to Cloud Storage
  supported_modes = ("data")
  if mode not in supported_modes:
    message = "Unsupported mode '%s', should be one of '%s'."
    raise ValueError(message, mode, supported_modes)

  image_data = serialize_array(array, fmt=fmt, quality=quality)
  base64_byte_string = base64.b64encode(image_data).decode('ascii')
  return "data:image/" + fmt.upper() + ";base64," + base64_byte_string


# public functions


def image(array, domain=None, w=None, format='png'):
  """Display an image.

  Args:
    array: NumPy array representing the image
    fmt: Image format e.g. png, jpeg
    domain: Domain of pixel values, inferred from min & max values if None
    w: width of output image, scaled using nearest neighbor interpolation.
      size unchanged if None
  """
  data_url = _image_url(array, domain=domain)
  html = '<img src=\"' + data_url + '\">'
  _display_html(html)


def images(arrays, labels=None, domain=None, w=None):
  """Display a list of images with optional labels.

  Args:
    arrays: A list of NumPy arrays representing images
    labels: A list of strings to label each image.
      Defaults to show index if None
    domain: Domain of pixel values, inferred from min & max values if None
    w: width of output image, scaled using nearest neighbor interpolation.
      size unchanged if None
  """

  s = '<div style="display: flex; flex-direction: row;">'
  for i, array in enumerate(arrays):
    url = _image_url(array)
    label = labels[i] if labels is not None else i
    s += """<div style="margin-right:10px;">
              {label}<br/>
              <img src="{url}" style="margin-top:4px;">
            </div>""".format(label=label, url=url)
  s += "</div>"
  _display_html(s)


def show(thing, domain=(0, 1)):
  """Display a nupmy array without having to specify what it represents.

  This module will attempt to infer how to display your tensor based on its
  rank, shape and dtype. rank 4 tensors will be displayed as image grids, rank
  2 and 3 tensors as images.
  """
  if isinstance(thing, np.ndarray):
    rank = len(thing.shape)
    if rank == 4:
      log.debug("Show is assuming rank 4 tensor to be a list of images.")
      images(thing, domain=domain)
    elif rank in (2, 3):
      log.debug("Show is assuming rank 2 or 3 tensor to be an image.")
      image(thing, domain=domain)
    else:
      log.warn("Show only supports numpy arrays of rank 2-4. Using repr().")
      print(repr(thing))
  elif isinstance(thing, (list, tuple)):
    log.debug("Show is assuming list or tuple to be a collection of images.")
    images(thing, domain=domain)
  else:
    log.warn("Show only supports numpy arrays so far. Using repr().")
    print(repr(thing))


def textured_mesh(mesh, texture, background='0xffffff'):
  texture_data_url = _image_url(texture, fmt='jpeg', quality=90)

  code = Template('''
  <input id="unfoldBox" type="checkbox" class="control">Unfold</input>
  <input id="shadeBox" type="checkbox" class="control">Shade</input>

  <script src="https://cdn.rawgit.com/mrdoob/three.js/r89/build/three.min.js"></script>
  <script src="https://cdn.rawgit.com/mrdoob/three.js/r89/examples/js/controls/OrbitControls.js"></script>

  <script type="x-shader/x-vertex" id="vertexShader">
    uniform float viewAspect;
    uniform float unfolding_perc;
    uniform float shadeFlag;
    varying vec2 text_coord;
    varying float shading;
    void main () {
      gl_Position = projectionMatrix * modelViewMatrix * vec4(position, 1.0);
      vec4 plane_position = vec4((uv.x*2.0-1.0)/viewAspect, (uv.y*2.0-1.0), 0, 1);
      gl_Position = mix(gl_Position, plane_position, unfolding_perc);

      //not normalized on purpose to simulate the rotation
      shading = 1.0;
      if (shadeFlag > 0.5) {
        vec3 light_vector = mix(normalize(cameraPosition-position), normal, unfolding_perc);
        shading = dot(normal, light_vector);
      }

      text_coord = uv;
    }
  </script>

  <script type="x-shader/x-fragment" id="fragmentShader">
    uniform float unfolding_perc;
    varying vec2  text_coord;
    varying float shading;
    uniform sampler2D texture;

    void main() {
      gl_FragColor = texture2D(texture, text_coord);
      gl_FragColor.rgb *= shading;
    }
  </script>

  <script>
  "use strict";

  const el = id => document.getElementById(id);

  const unfoldDuration = 1000.0;
  var camera, scene, renderer, controls, material;
  var unfolded = false;
  var unfoldStart = -unfoldDuration*10.0;

  init();
  animate(0.0);

  function init() {
    var width = 800, height = 600;

    scene = new THREE.Scene();

    camera = new THREE.PerspectiveCamera(50, width / height, 0.1, 100);
    camera.position.z = 1.5;
    scene.add(camera);

    controls = new THREE.OrbitControls( camera );

    var geometry = new THREE.BufferGeometry();
    geometry.addAttribute( 'position', new THREE.BufferAttribute($verts, 3 ) );
    geometry.addAttribute( 'uv', new THREE.BufferAttribute($uvs, 2) );
    geometry.setIndex(new THREE.BufferAttribute($faces, 1 ));
    geometry.computeVertexNormals();

    var texture = new THREE.TextureLoader().load('$tex_data_url', update);
    material = new THREE.ShaderMaterial( {
      uniforms: {
        viewAspect: {value: width/height},
        unfolding_perc: { value: 0.0 },
        shadeFlag: { value: 0.0 },
        texture: { type: 't', value: texture },
      },
      side: THREE.DoubleSide,
      vertexShader: el( 'vertexShader' ).textContent,
      fragmentShader: el( 'fragmentShader' ).textContent
    });

    var mesh = new THREE.Mesh(geometry, material);
    scene.add(mesh);
    scene.background = new THREE.Color( $background );

    renderer = new THREE.WebGLRenderer({antialias: true});
    renderer.setSize(width, height);
    document.body.appendChild(renderer.domElement);

    // render on change only
    controls.addEventListener('change', function() {
      // fold mesh back if user wants to interact
      el('unfoldBox').checked = false;
      update();
    });
    document.querySelectorAll('.control').forEach(e=>{
      e.addEventListener('change', update);
    });
  }

  function update() {
    requestAnimationFrame(animate);
  }

  function ease(x) {
    x = Math.min(Math.max(x, 0.0), 1.0);
    return x*x*(3.0 - 2.0*x);
  }

  function animate(time) {
    var unfoldFlag = el('unfoldBox').checked;
    if (unfolded != unfoldFlag) {
      unfolded = unfoldFlag;
      unfoldStart = time - Math.max(unfoldStart+unfoldDuration-time, 0.0);
    }
    var unfoldTime = (time-unfoldStart) / unfoldDuration;
    if (unfoldTime < 1.0) {
      update();
    }
    var unfoldVal = ease(unfoldTime);
    unfoldVal = unfolded ? unfoldVal : 1.0 - unfoldVal;
    material.uniforms.unfolding_perc.value = unfoldVal;

    material.uniforms.shadeFlag.value = el('shadeBox').checked ? 1.0 : 0.0;
    controls.update();
    renderer.render(scene, camera);
  }
  </script>
  ''').substitute(
      verts = array_to_jsbuffer(mesh['position'].ravel()),
      uvs = array_to_jsbuffer(mesh['uv'].ravel()),
      faces = array_to_jsbuffer(np.uint32(mesh['face'].ravel())),
      tex_data_url = texture_data_url,
      background = background,
  )
  _display_html(code)



"""Neural Image Style Transfer utils."""

import tensorflow as tf


def gram_style(a):
  with tf.name_scope('gram'):
    chn = int(a.shape[-1])
    a = tf.reshape(a, [-1, chn])
    n = tf.shape(a)[0]
    gram = tf.matmul(a, a, transpose_a=True)
    return gram / tf.cast(n, tf.float32)


def mean_l1_loss(g1, g2):
  with tf.name_scope('mean_l1_loss'):
    return tf.reduce_mean(tf.abs(g1-g2))


class StyleLoss(object):
  """Image Style Loss.
  
  A variant of style component of Artistic Image Style Transfer loss,
  mainly inspired by [1].

  [1] Leon A. Gatys et al. "A Neural Algorithm of Artistic Style"
      https://arxiv.org/abs/1508.06576
  """

  def __init__(self, style_layers, ema_decay=0.0,
               style_func=gram_style,
               loss_func=mean_l1_loss):
    """Initilize style loss.
    
    Args:
      style_layers: List of tensors that are used to compute statistics that
        define a style.
      ema_decay: Number in range [0.0 .. 1.0]. Loss function is computed against
        moving averaged versions of style statistics if ema_decay > 0.0.
        This is useful when each optimisation step only covers some part of
        the full output image.
      style_func: Function that is used to compute layer statistics.
      loss_func: Function that is used to compute difference between two
        outputs of 'style_func'.
    """
    self.input_grams = [style_func(s) for s in style_layers]
    if ema_decay:
      ema = tf.train.ExponentialMovingAverage(decay=ema_decay)
      update_ema_op = ema.apply(self.input_grams)
      with tf.control_dependencies([update_ema_op]):
        self.effective_grams = [g + tf.stop_gradient(ema.average(g)-g)
                           for g in self.input_grams]
    else:
      self.effective_grams = self.input_grams
    
    self.target_vars = [tf.Variable(tf.zeros_like(g), trainable=False)
                        for g in self.input_grams]
    self.style_losses = [loss_func(g, gv)
                         for g, gv in zip(self.effective_grams, self.target_vars)]
    self.style_loss = tf.add_n(self.style_losses)

  def set_style(self, input_feeds):
    """Set target style variables.
    
    Expected usage: 
      style_loss = StyleLoss(style_layers)
      ...
      init_op = tf.global_variables_initializer()
      init_op.run()
      
      feeds = {... session.run() 'feeds' argument that will make 'style_layers'
               tensors evaluate to activation values of style image...}
      style_loss.set_style(feeds)  # this must be called after 'init_op.run()'
    """
    sess = tf.get_default_session()
    computed = sess.run(self.input_grams, input_feeds)
    for v, g in zip(self.target_vars, computed):
      v.load(g)



"""OpenGL Mesh rendering utils."""

from contextlib import contextmanager
import numpy as np

import OpenGL.GL as gl

from .meshutil import perspective


class GLObject(object):
  def __del__(self):
    self.release()
  def __enter__(self):
    bind_func, const = self._bind
    bind_func(const, self)
  def __exit__(self, *args):
    bind_func, const = self._bind
    bind_func(const, 0)
    
class FBO(GLObject):
  _bind = gl.glBindFramebuffer, gl.GL_FRAMEBUFFER
  def __init__(self):
    self._as_parameter_ = gl.glGenFramebuffers(1)
  def release(self):
    gl.glDeleteFramebuffers(1, [self._as_parameter_])
    

class Texture(GLObject):
  _bind = gl.glBindTexture, gl.GL_TEXTURE_2D
  def __init__(self):
    self._as_parameter_ = gl.glGenTextures(1)
  def release(self):
    gl.glDeleteTextures([self._as_parameter_])
    

class Shader(GLObject):

  def __init__(self, vp_code, fp_code):
    # Importing here, when gl context is already present.
    # Otherwise get expection on Python3 because of PyOpenGL bug.
    from OpenGL.GL import shaders
    self._as_parameter_ = self._shader = shaders.compileProgram(
        shaders.compileShader( vp_code, gl.GL_VERTEX_SHADER ),
        shaders.compileShader( fp_code, gl.GL_FRAGMENT_SHADER )
    )
    self._uniforms = {}
    
  def release(self):
    gl.glDeleteProgram(self._as_parameter_)
  
  def __getitem__(self, uniform_name):
    if uniform_name not in self._uniforms:
      self._uniforms[uniform_name] = gl.glGetUniformLocation(self, uniform_name)
    return self._uniforms[uniform_name]

  def __enter__(self):
    return self._shader.__enter__()
  def __exit__(self, *args):
    return self._shader.__exit__(*args)
      
    
class MeshRenderer(object):
  def __init__(self, size):
    self.size = size
    self.fbo = FBO()
    self.color_tex = Texture()
    self.depth_tex = Texture()
    w, h = size
    
    with self.color_tex:
      gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA32F, w, h, 0,
                      gl.GL_RGBA, gl.GL_FLOAT, None)
      
    with self.depth_tex:
      gl.glTexImage2D.wrappedOperation(
          gl.GL_TEXTURE_2D, 0,gl.GL_DEPTH24_STENCIL8, w, h, 0,
          gl.GL_DEPTH_STENCIL, gl.GL_UNSIGNED_INT_24_8, None)
    
    with self.fbo:
      gl.glFramebufferTexture2D(gl.GL_FRAMEBUFFER, gl.GL_COLOR_ATTACHMENT0,
                             gl.GL_TEXTURE_2D, self.color_tex, 0)
      gl.glFramebufferTexture2D(gl.GL_FRAMEBUFFER, gl.GL_DEPTH_STENCIL_ATTACHMENT,
                             gl.GL_TEXTURE_2D, self.depth_tex, 0)
      gl.glViewport(0, 0, w, h)
      assert gl.glCheckFramebufferStatus(gl.GL_FRAMEBUFFER) == gl.GL_FRAMEBUFFER_COMPLETE
        
    self.shader = Shader(vp_code='''
      #version 130
      uniform mat4 MVP;
      in vec4 data;
      out vec4 aData;

      void main() {
        aData = data;
        gl_Position = MVP * gl_Vertex;
      }
    ''', 
    fp_code='''
      #version 130
      in vec4 aData;
      out vec4 fragColor;
      void main() {
        fragColor = aData;
      }
    ''')
    
    self.fovy = 10.0
    self.aspect = 1.0*w/h
    self.znear, self.zfar = 0.01, 100.0
    
  @contextmanager
  def _bind_attrib(self, i, arr):
    if arr is None:
      yield
      return
    arr = np.ascontiguousarray(arr, np.float32)
    coord_n = arr.shape[-1]
    gl.glEnableVertexAttribArray(i)
    gl.glVertexAttribPointer(i, coord_n, gl.GL_FLOAT, gl.GL_FALSE, 0, arr)
    yield
    gl.glDisableVertexAttribArray(i)
    
  def proj_matrix(self):
    return perspective(self.fovy, self.aspect, self.znear, self.zfar)
    
  def render_mesh(self, position, uv, face=None,
                  clear_color=[0, 0, 0, 0],
                  modelview=np.eye(4)):
    MVP = modelview.T.dot(self.proj_matrix())
    MVP = np.ascontiguousarray(MVP, np.float32)
    position = np.ascontiguousarray(position, np.float32)
    with self.fbo:
      gl.glClearColor(*clear_color)
      gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
      
      with self.shader, self._bind_attrib(0, position), self._bind_attrib(1, uv):
        gl.glUniformMatrix4fv(self.shader['MVP'], 1, gl.GL_FALSE, MVP)
        gl.glEnable(gl.GL_DEPTH_TEST)
        if face is not None:
          face = np.ascontiguousarray(face, np.uint32)
          gl.glDrawElements(gl.GL_TRIANGLES, face.size, gl.GL_UNSIGNED_INT, face)
        else:
          vert_n = position.size//position.shape[-1]
          gl.glDrawArrays(gl.GL_TRIANGLES, 0, vert_n)
        gl.glDisable(gl.GL_DEPTH_TEST)
      
      w, h = self.size
      frame = gl.glReadPixels(0, 0, w, h, gl.GL_RGBA, gl.GL_FLOAT)
      frame = frame.reshape(h, w, 4)  # fix PyOpenGL bug
      frame = frame[::-1]  # verical flip to match GL convention
      return frame


# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from __future__ import absolute_import, division, print_function

import tensorflow as tf
from lucid.modelzoo.vision_base import Model


def populate_inception_bottlenecks(scope):
  """Add Inception bottlenecks and their pre-Relu versions to the graph."""
  graph = tf.get_default_graph()
  for op in graph.get_operations():
    if op.name.startswith(scope+'/') and 'Concat' in op.type:
      name = op.name.split('/')[1]
      pre_relus = []
      for tower in op.inputs[1:]:
        if tower.op.type == 'Relu':
          tower = tower.op.inputs[0]
        pre_relus.append(tower)
      concat_name = scope + '/' + name + '_pre_relu'
      _ = tf.concat(pre_relus, -1, name=concat_name)


class InceptionV1(Model):
  model_path = 'gs://modelzoo/InceptionV1.pb'
  labels_path = 'gs://modelzoo/InceptionV1-labels.txt'
  image_shape = [224, 224, 3]
  image_value_range = (-117, 255-117)
  input_name = 'input:0'

  def post_import(self, scope):
    populate_inception_bottlenecks(scope)



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Functions for transforming and constraining color channels."""


import numpy as np
import tensorflow as tf

from lucid.optvis.param.unit_balls import constrain_L_inf

color_correlation_svd_sqrt = np.asarray([[0.26, 0.09, 0.02],
                                         [0.27, 0.00, -0.05],
                                         [0.27, -0.09, 0.03]]).astype("float32")
max_norm_svd_sqrt = np.max(np.linalg.norm(color_correlation_svd_sqrt, axis=0))

color_mean = [0.48, 0.46, 0.41]       


def _linear_decorelate_color(t):
  """Multiply input by sqrt of emperical (ImageNet) color correlation matrix.
  
  If you interpret t's innermost dimension as describing colors in a
  decorrelated version of the color space (which is a very natural way to
  describe colors -- see discussion in Feature Visualization article) the way
  to map back to normal colors is multiply the square root of your color
  correlations. 
  """
  # check that inner dimension is 3?
  t_flat = tf.reshape(t, [-1, 3])
  color_correlation_normalized = color_correlation_svd_sqrt / max_norm_svd_sqrt
  t_flat = tf.matmul(t_flat, color_correlation_normalized.T)
  t = tf.reshape(t_flat, tf.shape(t))
  return t


def to_valid_rgb(t, decorrelate=False, sigmoid=True):
  """Transform inner dimension of t to valid rgb colors.
  
  In practice this consistes of two parts: 
  (1) If requested, transform the colors from a decorrelated color space to RGB.
  (2) Constrain the color channels to be in [0,1], either using a sigmoid
      function or clipping.
  
  Args:
    t: input tensor, innermost dimension will be interpreted as colors
      and transformed/constrained.
    decorrelate: should the input tensor's colors be interpreted as coming from
      a whitened space or not?
    sigmoid: should the colors be constrained using sigmoid (if True) or
      clipping (if False).
  
  Returns:
    t with the innermost dimension transformed.
  """
  if decorrelate:
    t = _linear_decorelate_color(t)
  if decorrelate and not sigmoid:
    t += color_mean
  if sigmoid:
    return tf.nn.sigmoid(t)
  else:
    return constrain_L_inf(2*t-1)/2 + 0.5



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Method for saving arbitrary data to arbitrary destinations.

This module takes an object and URL, infers how to serialize and how to write
it out to the destination.

If an object could have multiple representations, this tries to infer the
intended representation from the URL's file extension.

Possible extension: if not given a URL this could create one and return it?
"""

from __future__ import absolute_import, division, print_function

import logging
import os.path
import json
import numpy as np
import PIL.Image

from lucid.misc.io.writing import write, write_handle
from lucid.misc.io.serialize_array import _normalize_array


# create logger with module name, e.g. lucid.misc.io.saving
log = logging.getLogger(__name__)


def save_json(object, handle, indent=2):
  """Save object as json on CNS."""
  obj_json = json.dumps(object, indent=indent)
  handle.write(obj_json)


def save_npy(object, handle):
  """Save numpy array as npy file."""
  np.save(handle, object)


def save_npz(object, handle):
  """Save dict of numpy array as npz file."""
  # there is a bug where savez doesn't actually accept a file handle.
  log.warn("Saving npz files currently only works locally. :/")
  path = handle.name
  handle.close()
  if type(object) is dict:
    np.savez(path, **object)
  elif type(object) is list:
    np.savez(path, *object)
  else:
    log.warn("Saving non dict or list as npz file, did you maybe want npy?")
    np.savez(path, object)


def save_img(object, handle, **kwargs):
  """Save numpy array as image file on CNS."""
  if isinstance(object, np.ndarray):
    normalized = _normalize_array(object)
    image = PIL.Image.fromarray(normalized)
  elif isinstance(object, PIL.Image):
    image = object
  else:
    raise ValueError("Can only save_img for numpy arrays or PIL.Images!")
  image.save(handle, **kwargs)  # will infer format from handle's url ext.


savers = {
  ".png": save_img,
  ".jpg": save_img,
  ".jpeg": save_img,
  ".npy": save_npy,
  ".npz": save_npz,
  ".json": save_json,
}


def save(thing, url_or_handle, **kwargs):
  """Save object to file on CNS.

  File format is inferred from path. Use save_img(), save_npy(), or save_json()
  if you need to force a particular format.

  Args:
    obj: object to save.
    path: CNS path.

  Raises:
    RuntimeError: If file extension not supported.
  """
  is_handle = hasattr(url_or_handle, 'write') and hasattr(url_or_handle, 'name')
  if is_handle:
    _, ext = os.path.splitext(url_or_handle.name)
  else:
    _, ext = os.path.splitext(url_or_handle)
  if not ext:
    raise RuntimeError("No extension in URL: " + url_or_handle)

  if ext in savers:
    saver = savers[ext]
    if is_handle:
      saver(thing, url_or_handle, **kwargs)
    else:
      with write_handle(url_or_handle) as handle:
        saver(thing, handle, **kwargs)
  else:
    message = "Unknown extension '{}', supports {}."
    raise RuntimeError(message.format(ext, savers))



"""Optimize within unit balls in TensorFlow.

In adverserial examples, one often wants to optize within a constrained ball.
This module makes this easy through functions like unit_ball_L2(), which
creates a tensorflow variable constrained within a L2 unit ball.

EXPERIMENTAL: Do not use for adverserial examples if you need to be confident
they are strong attacks. We are not yet confident in this code.
"""

import tensorflow as tf

from lucid.misc.gradient_override import use_gradient


def dot(a, b):
  return tf.reduce_sum(a * b)


def _constrain_L2_grad(op, grad):
  """Gradient for constrained optimization on an L2 unit ball.

  This function projects the gradient onto the ball if you are on the boundary
  (or outside!), but leaves it untouched if you are inside the ball.

  Args:
    op: the tensorflow op we're computing the gradient for.
    grad: gradient we need to backprop

  Returns:
    (projected if necessary) gradient.
  """
  inp = op.inputs[0]
  inp_norm = tf.norm(inp)
  unit_inp = inp / inp_norm

  grad_projection = dot(unit_inp, grad)
  parallel_grad = unit_inp * grad_projection

  is_in_ball = tf.less_equal(inp_norm, 1)
  is_pointed_inward = tf.less(grad_projection, 0)
  allow_grad = tf.logical_or(is_in_ball, is_pointed_inward)
  clip_grad = tf.logical_not(allow_grad)

  clipped_grad = tf.cond(clip_grad, lambda: grad - parallel_grad, lambda: grad)

  return clipped_grad


@use_gradient(_constrain_L2_grad)
def constrain_L2(x):
  return x / tf.maximum(1.0, tf.norm(x))


def unit_ball_L2(shape):
  """A tensorflow variable tranfomed to be constrained in a L2 unit ball.

  EXPERIMENTAL: Do not use for adverserial examples if you need to be confident
  they are strong attacks. We are not yet confident in this code.
  """
  x = tf.Variable(tf.zeros(shape))
  return constrain_L2(x)


def _constrain_L_inf_grad(precondition=True):

  def grad_f(op, grad):
    """Gradient for constrained preconditioned optimization on an L_inf unit
    ball.

    This function projects the gradient onto the ball if you are on the
    boundary (or outside!). It always preconditions the gradient so it is the
    direction of steepest descent under L_inf.

    Args:
      op: the tensorflow op we're computing the gradient for.
      grad: gradient we need to backprop

    Returns:
      (projected if necessary) preconditioned gradient.
    """
    inp = op.inputs[0]
    dim_at_edge = tf.greater_equal(tf.abs(inp), 1.0)
    dim_outward = tf.greater(inp * grad, 0.0)
    if precondition:
      grad = tf.sign(grad)

    return tf.where(
        tf.logical_and(dim_at_edge, dim_outward),
        tf.zeros(grad.shape),
        grad
      )
  return grad_f


@use_gradient(_constrain_L_inf_grad(precondition=True))
def constrain_L_inf_precondition(x):
  return x / tf.maximum(1.0, tf.abs(x))


@use_gradient(_constrain_L_inf_grad(precondition=False))
def constrain_L_inf(x):
  return x / tf.maximum(1.0, tf.abs(x))


def unit_ball_L_inf(shape, precondition=True):
  """A tensorflow variable tranfomed to be constrained in a L_inf unit ball.

  Note that this code also preconditions the gradient to go in the L_inf
  direction of steepest descent.

  EXPERIMENTAL: Do not use for adverserial examples if you need to be confident
  they are strong attacks. We are not yet confident in this code.
  """
  x = tf.Variable(tf.zeros(shape))
  if precondition:
    return constrain_L_inf_precondition(x)
  else:
    return constrain_L_inf(x)



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Provides render_vis() for actually rendering visualizations.

This module primarily provides render_vis() for rendering visualizations.
It also provides some utilities in case you need to create your own rendering
function.
"""
from __future__ import absolute_import, division, print_function
from future.standard_library import install_aliases
install_aliases()
from builtins import range

import numpy as np
import tensorflow as tf
import logging

from lucid.optvis import objectives, param, transform
from lucid.misc.io import show

# pylint: disable=invalid-name


# create logger with module name, e.g. lucid.misc.io.reading
log = logging.getLogger(__name__)


def render_vis(model, objective_f, param_f=None, optimizer=None,
               transforms=None, thresholds=(512,),
               print_objectives=None, verbose=True,):
  """Flexible optimization-base feature vis.

  There's a lot of ways one might wish to customize otpimization-based
  feature visualization. It's hard to create an abstraction that stands up
  to all the things one might wish to try.

  This function probably can't do *everything* you want, but it's much more
  flexible than a naive attempt. The basic abstraction is to split the problem
  into several parts. Consider the rguments:

  Args:
    model: The model to be visualized, from Alex' modelzoo.
    objective_f: The objective our visualization maximizes.
      See the objectives module for more details.
    param_f: Paramaterization of the image we're optimizing.
      See the paramaterization module for more details.
      Defaults to a naively paramaterized [1, 128, 128, 3] image.
    optimizer: Optimizer to optimize with. Either tf.train.Optimizer instance,
      or a function from (graph, sess) to such an instance.
      Defaults to Adam with lr .05.
    transforms: A list of stochastic transformations that get composed,
      which our visualization should robustly activate the network against.
      See the transform module for more details.
      Defaults to [transform.jitter(8)].
    thresholds: A list of numbers of optimization steps, at which we should
      save (and display if verbose=True) the visualization.
    print_objectives: A list of objectives separate from those being optimized,
      whose values get logged during the optimization.
    verbose: Should we display the visualization when we hit a threshold?
      This should only be used in IPython.
  Returns:
    2D array of optimization results containing of evaluations of supplied
    param_f snapshotted at specified thresholds. Usually that will mean one or
    multiple channel visualizations stacked on top of each other.
  """

  with tf.Graph().as_default() as graph, tf.Session() as sess:

    T = make_vis_T(model, objective_f, param_f, optimizer, transforms)
    print_objective_func = make_print_objective_func(print_objectives, T)
    loss, vis_op, t_image = T("loss"), T("vis_op"), T("input")
    tf.global_variables_initializer().run()

    images = []
    try:
      for i in range(max(thresholds)+1):
        loss_, _ = sess.run([loss, vis_op])
        if i in thresholds:
          vis = t_image.eval()
          images.append(vis)
          if verbose:
            print(i, loss_)
            print_objective_func(sess)
            show(np.hstack(vis))
    except KeyboardInterrupt:
      log.warn("Interrupted optimization at step {:d}.".format(i+1))
      vis = t_image.eval()
      show(np.hstack(vis))

    return images


def make_vis_T(model, objective_f, param_f=None, optimizer=None,
               transforms=None):
  """Even more flexible optimization-base feature vis.

  This function is the inner core of render_vis(), and can be used
  when render_vis() isn't flexible enough. Unfortunately, it's a bit more
  tedious to use:

  >  with tf.Graph().as_default() as graph, tf.Session() as sess:
  >
  >    T = make_vis_T(model, "mixed4a_pre_relu:0")
  >    tf.initialize_all_variables().run()
  >
  >    for i in range(10):
  >      T("vis_op").run()
  >      showarray(T("input").eval()[0])

  This approach allows more control over how the visualizaiton is displayed
  as it renders. It also allows a lot more flexibility in constructing
  objectives / params because the session is already in scope.


  Args:
    model: The model to be visualized, from Alex' modelzoo.
    objective_f: The objective our visualization maximizes.
      See the objectives module for more details.
    param_f: Paramaterization of the image we're optimizing.
      See the paramaterization module for more details.
      Defaults to a naively paramaterized [1, 128, 128, 3] image.
    optimizer: Optimizer to optimize with. Either tf.train.Optimizer instance,
      or a function from (graph, sess) to such an instance.
      Defaults to Adam with lr .05.
    transforms: A list of stochastic transformations that get composed,
      which our visualization should robustly activate the network against.
      See the transform module for more details.
      Defaults to [transform.jitter(8)].

  Returns:
    A function T, which allows access to:
      * T("vis_op") -- the operation for to optimize the visualization
      * T("input") -- the visualization itself
      * T("loss") -- the loss for the visualization
      * T(layer) -- any layer inside the network
  """

  # pylint: disable=unused-variable
  t_image = make_t_image(param_f)
  objective_f = objectives.as_objective(objective_f)
  transform_f = make_transform_f(transforms)
  optimizer = make_optimizer(optimizer, [])

  T = import_model(model, transform_f(t_image), t_image)
  loss = objective_f(T)

  global_step = tf.Variable(0, trainable=False, name="global_step")
  vis_op = optimizer.minimize(-loss, global_step=global_step)

  local_vars = locals()
  # pylint: enable=unused-variable

  def T2(name):
    if name in local_vars:
      return local_vars[name]
    else: return T(name)

  return T2


def make_print_objective_func(print_objectives, T):
  print_objectives = print_objectives or []
  po_descriptions = [obj.description for obj in print_objectives]
  pos = [obj(T) for obj in print_objectives]

  def print_objective_func(sess):
    pos_results = sess.run(pos)
    for k, v, i in zip(po_descriptions, pos_results, range(len(pos_results))):
      print("{:02d}: {}: {:7.2f}".format(i+1, k, v))

  return print_objective_func

# pylint: enable=invalid-name


def make_t_image(param_f):
  if param_f is None:
    t_image = param.image(128)
  elif callable(param_f):
    t_image = param_f()
  elif isinstance(param_f, tf.Tensor):
    t_image = param_f
  else:
    raise TypeError("Incompatible type for param_f, " + str(type(param_f)) )

  if not isinstance(t_image, tf.Tensor):
    raise TypeError("param_f should produce a Tensor, but instead created a "
                   + str(type(t_image)) )
  elif t_image.graph != tf.get_default_graph():
    raise TypeError("""param_f produced a t_image tensor belonging to a graph
                     that isn't the default graph for rendering. Did you
                     accidentally use render_vis when you meant to use
                     make_vis_T?""")
  else:
    return t_image


def make_transform_f(transforms):
  if type(transforms) is not list:
    transforms = transform.standard_transforms
  transform_f = transform.compose(transforms)
  return transform_f


def make_optimizer(optimizer, args):
  if optimizer is None:
    return tf.train.AdamOptimizer(0.05)
  elif callable(optimizer):
    return optimizer(*args)
  elif isinstance(optimizer, tf.train.Optimizer):
    return optimizer
  else:
    raise ("Could not convert optimizer argument to usable optimizer. "
           "Needs to be one of None, function from (graph, sess) to "
           "optimizer, or tf.train.Optimizer instance.")


def import_model(model, t_image, t_image_raw):

  model.import_graph(t_image, scope="import", forget_xy_shape=True)

  def T(layer):
    if layer == "input": return t_image_raw
    if layer == "labels": return model.labels
    return t_image.graph.get_tensor_by_name("import/%s:0"%layer)

  return T



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Provides resize_bilinear_nd.

This module provides resize_bilinear_nd, a function for resizing a tensor
with bilinear interpolation in n dimesions. It iteratively
applies tf.image.resize_bilinear (which can only resize 2 dimensions).
"""

import tensorflow as tf


def product(l):
  """Multiply together the elements of a list."""
  prod = 1
  for x in l:
    prod *= x
  return prod


def collapse_shape(shape, a, b):
  """Collapse `shape` outside the interval (`a`,`b`).

  This function collapses `shape` outside the interval (`a`,`b`) by
  multiplying the dimensions before `a` into a single dimension,
  and mutliplying the dimensions after `b` into a single dimension.

  Args:
    shape: a tensor shape
    a: integer, position in shape
    b: integer, position in shape

  Returns:
    The collapsed shape, represented as a list.

  Examples:
    [1, 2, 3, 4, 5], (a=0, b=2) => [1, 1, 2, 60]
    [1, 2, 3, 4, 5], (a=1, b=3) => [1, 2, 3, 20]
    [1, 2, 3, 4, 5], (a=2, b=4) => [2, 3, 4, 5 ]
    [1, 2, 3, 4, 5], (a=3, b=5) => [6, 4, 5, 1 ]
  """
  shape = list(shape)
  if a < 0:
    n_pad = -a
    pad = n_pad * [1]
    return collapse_shape(pad + shape, a + n_pad, b + n_pad)
  if b > len(shape):
    n_pad = b - len(shape)
    pad = n_pad * [1]
    return collapse_shape(shape + pad, a, b)
  return [product(shape[:a])] + shape[a:b] + [product(shape[b:])]


def resize_bilinear_nd(t, target_shape):
  """Bilinear resizes a tensor t to have shape target_shape.

  This function bilinearly resizes a n-dimensional tensor by iteratively
  applying tf.image.resize_bilinear (which can only resize 2 dimensions).
  For bilinear interpolation, the order in which it is applied does not matter.

  Args:
    t: tensor to be resized
    target_shape: the desired shape of the new tensor.

  Returns:
   The resized tensor
  """
  shape = t.get_shape().as_list()
  target_shape = list(target_shape)
  assert len(shape) == len(target_shape)

  # We progressively move through the shape, resizing dimensions...
  d = 0
  while d < len(shape):

    # If we don't need to deal with the next dimesnion, step over it
    if shape[d] == target_shape[d]:
      d += 1
      continue

    # Otherwise, we'll resize the next two dimensions...
    # If d+2 doesn't need to be resized, this will just be a null op for it
    new_shape = shape[:]
    new_shape[d : d+2] = target_shape[d : d+2]

    # The helper collapse_shape() makes our shapes 4-dimensional with
    # the two dimesnions we want to deal with in the middle.
    shape_ = collapse_shape(shape, d, d+2)
    new_shape_ = collapse_shape(new_shape, d, d+2)

    # We can then reshape and use the 2d tf.image.resize_bilinear() on the
    # inner two dimesions.
    t_ = tf.reshape(t, shape_)
    t_ = tf.image.resize_bilinear(t_, new_shape_[1:3])

    # And then reshape back to our uncollapsed version, having finished resizing
    # two more dimensions in our shape.
    t = tf.reshape(t_, new_shape)
    shape = new_shape
    d += 2

  return t



""" Logging in lucid

Lucid uses Python's default logging system.
Each module has it's own logger, so you can control the verbosity of each
module to your needs. To change the overall log level, you can set log levels
at each part of the module hierarchy, including simply at the root, `lucid`:

```python
import logging
logging.getLogger('lucid').setLevel(logging.INFO)
```

Our log levels also confirm to Python's defaults:

DEBUG	Detailed information typically of interest only when diagnosing problems.
INFO	Confirmation that things are working as expected.
WARN	An indication that something unexpected happened.
ERROR	The software has not been able to perform some function.
CRITICAL	The program itself may be unable to continue running.

"""

import logging
logging.basicConfig(level=logging.WARN)



# Copyright 2017 The TensorFlow Lucid Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Setup script for TensorFlow Lucid."""

from __future__ import absolute_import, division, print_function

from setuptools import setup, find_packages

version = '0.0.8'

test_deps = [
  'future',
  'twine',
  'pytest',
  'pytest-mock',
  'python-coveralls'
]

extras = {
  'test': test_deps,
}

setup(
  name = 'lucid',
  packages = find_packages(exclude=[]),
  version = version,
  description = ('Collection of infrastructure and tools for research in '
    'neural network interpretability.'),
  author = 'The Lucid Authors',
  author_email = 'deepviz@google.com',
  url = 'https://github.com/tensorflow/lucid',
  download_url = ('https://github.com/tensorflow/lucid'
    '/archive/v{}.tar.gz'.format(version)),
  license = 'Apache License 2.0',
  keywords = ['tensorflow', 'tensor', 'machine learning', 'neural networks',
    'convolutional neural networks', 'feature visualization', 'optimization'],
  install_requires = [
    'future',
    'decorator',
    'tensorflow',
    'scikit-learn',
    'numpy',
    'scipy',
    'pillow',
    'pyopengl',
    'ipython'
  ],
  tests_require = test_deps,
  extras_require = extras,
  classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Education',
    'License :: OSI Approved :: Apache Software License',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Visualization',
    'Topic :: Software Development :: Libraries :: Python Modules',
  ],
)



"""Headless GPU-accelerated OpenGL context creation on Google Colaboratory.

Typical usage:

    # Optional PyOpenGL configuratiopn can be done here.
    # import OpenGL
    # OpenGL.ERROR_CHECKING = True

    # 'glcontext' must be imported before any OpenGL.* API.
    from lucid.misc.gl.glcontext import create_opengl_context

    # Now it's safe to import OpenGL and EGL functions
    import OpenGL.GL as gl

    # create_opengl_context() creates a GL context that is attached to an
    # offscreen surface of the specified size. Note that rendering to buffers
    # of other sizes and formats is still possible with OpenGL Framebuffers.
    #
    # Users are expected to directly use the EGL API in case more advanced
    # context management is required.
    width, height = 640, 480
    create_opengl_context((width, height))

    # OpenGL context is available here.

"""

from __future__ import print_function

# pylint: disable=unused-import,g-import-not-at-top,g-statement-before-imports

try:
  import OpenGL
except:
  print('This module depends on PyOpenGL.')
  print('Please run "\033[1m!pip install -q pyopengl\033[0m" '
        'prior importing this module.')
  raise

import ctypes
from ctypes import pointer
import os

os.environ['PYOPENGL_PLATFORM'] = 'egl'

# OpenGL loading workaround.
#
# * PyOpenGL tries to load libGL, but we need libOpenGL, see [1,2].
#   This could have been solved by a symlink libGL->libOpenGL, but:
#
# * Python 2.7 can't find libGL and linEGL due to a bug (see [3])
#   in ctypes.util, that was only wixed in Python 3.6.
#
# So, the only solution I've found is to monkeypatch ctypes.util
# [1] https://devblogs.nvidia.com/egl-eye-opengl-visualization-without-x-server/
# [2] https://devblogs.nvidia.com/linking-opengl-server-side-rendering/
# [3] https://bugs.python.org/issue9998
_find_library_old = ctypes.util.find_library
try:

  def _find_library_new(name):
    return {
        'GL': 'libOpenGL.so',
        'EGL': 'libEGL.so',
    }.get(name, _find_library_old(name))
  ctypes.util.find_library = _find_library_new
  import OpenGL.GL as gl
  import OpenGL.EGL as egl
except:
  print('Unable to load OpenGL libraries. '
        'Make sure you use GPU-enabled backend.')
  print('Press "Runtime->Change runtime type" and set '
        '"Hardware accelerator" to GPU.')
  raise
finally:
  ctypes.util.find_library = _find_library_old


def create_opengl_context(surface_size=(640, 480)):
  """Create offscreen OpenGL context and make it current.

  Users are expected to directly use EGL API in case more advanced
  context management is required.

  Args:
    surface_size: (width, height), size of the offscreen rendering surface.
  """
  egl_display = egl.eglGetDisplay(egl.EGL_DEFAULT_DISPLAY)

  major, minor = egl.EGLint(), egl.EGLint()
  egl.eglInitialize(egl_display, pointer(major), pointer(minor))

  config_attribs = [
      egl.EGL_SURFACE_TYPE, egl.EGL_PBUFFER_BIT, egl.EGL_BLUE_SIZE, 8,
      egl.EGL_GREEN_SIZE, 8, egl.EGL_RED_SIZE, 8, egl.EGL_DEPTH_SIZE, 24,
      egl.EGL_RENDERABLE_TYPE, egl.EGL_OPENGL_BIT, egl.EGL_NONE
  ]
  config_attribs = (egl.EGLint * len(config_attribs))(*config_attribs)

  num_configs = egl.EGLint()
  egl_cfg = egl.EGLConfig()
  egl.eglChooseConfig(egl_display, config_attribs, pointer(egl_cfg), 1,
                      pointer(num_configs))

  width, height = surface_size
  pbuffer_attribs = [
      egl.EGL_WIDTH,
      width,
      egl.EGL_HEIGHT,
      height,
      egl.EGL_NONE,
  ]
  pbuffer_attribs = (egl.EGLint * len(pbuffer_attribs))(*pbuffer_attribs)
  egl_surf = egl.eglCreatePbufferSurface(egl_display, egl_cfg, pbuffer_attribs)

  egl.eglBindAPI(egl.EGL_OPENGL_API)

  egl_context = egl.eglCreateContext(egl_display, egl_cfg, egl.EGL_NO_CONTEXT,
                                     None)
  egl.eglMakeCurrent(egl_display, egl_surf, egl_surf, egl_context)


# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions for modelzoo models."""

from __future__ import absolute_import, division, print_function

import tensorflow as tf
from lucid.misc.io.reading import read

def load_text_labels(labels_path):
  return read(labels_path, encoding='utf-8').splitlines()

def load_graphdef(model_url, reset_device=True):
  """Load GraphDef from a binary proto file."""
  graphdef_string = read(model_url)
  graph_def = tf.GraphDef.FromString(graphdef_string)
  if reset_device:
    for n in graph_def.node:
      n.device = ""
  return graph_def

def forget_xy(t):
  """Forget sizes of dimensions [1, 2] of a 4d tensor."""
  shape = (t.shape[0], None, None, t.shape[3])
  return tf.placeholder_with_default(t, shape)



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Provides lowres_tensor()."""

import numpy as np
import tensorflow as tf

from lucid.optvis.param.resize_bilinear_nd import resize_bilinear_nd



def lowres_tensor(shape, underlying_shape, offset=None, sd=None):
  """Produces a tensor paramaterized by a interpolated lower resolution tensor.

  This is like what is done in a laplacian pyramid, but a bit more general. It
  can be a powerful way to describe images.

  Args:
    shape: desired shape of resulting tensor
    underlying_shape: shape of the tensor being resized into final tensor
    offset: Describes how to offset the interpolated vector (like phase in a
      Fourier transform). If None, apply no offset. If a scalar, apply the same
      offset to each dimension; if a list use each entry for each dimension.
      If a int, offset by that much. If False, do not offset. If True, offset by
      half the ratio between shape and underlying shape (analagous to 90
      degrees).
    sd: Standard deviation of initial tensor variable.

  Returns:
    A tensor paramaterized by a lower resolution tensorflow variable.
  """
  sd = sd or 0.01
  init_val = sd*np.random.randn(*underlying_shape).astype("float32")
  underlying_t = tf.Variable(init_val)
  t = resize_bilinear_nd(underlying_t, shape)
  if offset is not None:
    # Deal with non-list offset
    if not isinstance(offset, list):
      offset = len(shape)*[offset]
    # Deal with the non-int offset entries
    for n in range(len(offset)):
      if offset[n] is True:
        offset[n] = shape[n]/underlying_shape[n]/2
      if offset[n] is False:
        offset[n] = 0
      offset[n] = int(offset[n])
    # Actually apply offset by padding and then croping off the excess.
    padding = [(pad, 0) for pad in offset]
    t = tf.pad(t, padding, "SYMMETRIC")
    begin = len(shape)*[0]
    t = tf.slice(t, begin, shape)
  return t



# Copyright 2018 The Lucid Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


"""Tranformations you might want neural net visualizations to be robust to.

This module provides a variety of functions which stochastically transform a
tensorflow tensor. The functions are of the form:

  (config) => (tensor) => (stochastic transformed tensor)

"""

import tensorflow as tf
import numpy as np
import uuid

from lucid.optvis import param


def jitter(d, seed=None):
  def inner(t_image):
    t_image = tf.convert_to_tensor(t_image, preferred_dtype=tf.float32)
    t_shp = tf.shape(t_image)
    crop_shape = tf.concat([t_shp[:-3], t_shp[-3:-1] - d, t_shp[-1:]], 0)
    crop = tf.random_crop(t_image, crop_shape, seed=seed)
    shp = t_image.get_shape().as_list()
    mid_shp_changed = [shp[-3] - d if shp[-3] is not None else None,
                       shp[-2] - d if shp[-3] is not None else None]
    crop.set_shape(shp[:-3] + mid_shp_changed + shp[-1:])
    return crop
  return inner


def pad(w, mode="REFLECT", constant_value=0.5):
  def inner(t_image):
    if constant_value == "uniform":
      constant_value_ = tf.random_uniform([], 0, 1)
    else:
      constant_value_ = constant_value
    return tf.pad(t_image, [(0,0), (w,w), (w,w), (0,0)], mode=mode, constant_values=constant_value_)
  return inner


# def random_scale(scales, seed=None):
#   def inner(t):
#     t = tf.convert_to_tensor(t, preferred_dtype="float32")
#     scale = _rand_select(scales, seed=seed)
#     shp = tf.shape(t)
#     scale_shape = tf.concat(
#         [shp[:-3], tf.cast(scale * tf.cast(shp[-3:-1], "float32"), "int32"), shp[-1:]], 0)
#     return resize_bilinear_nd(t, scale_shape)
#   return inner

# 2D only version
def random_scale(scales, seed=None):
  def inner(t):
    t = tf.convert_to_tensor(t, preferred_dtype=tf.float32)
    scale = _rand_select(scales, seed=seed)
    shp = tf.shape(t)
    scale_shape = tf.cast(scale * tf.cast(shp[-3:-1], "float32"), "int32")
    return tf.image.resize_bilinear(t, scale_shape)
  return inner


def random_rotate(angles, units="degrees", seed=None):
  def inner(t):
    t = tf.convert_to_tensor(t, preferred_dtype=tf.float32)
    angle = _rand_select(angles, seed=seed)
    angle = _angle2rads(angle, units)
    return tf.contrib.image.rotate(t, angle)
  return inner


def normalize_gradient(grad_scales=None):

  if grad_scales is not None:
    grad_scales = np.float32(grad_scales)

  op_name = "NormalizeGrad_" + str(uuid.uuid4())
  @tf.RegisterGradient(op_name)
  def _NormalizeGrad(op, grad):
    grad_norm = tf.sqrt(tf.reduce_sum(grad**2, [1, 2, 3], keep_dims=True))
    if grad_scales is not None:
      grad *= grad_scales[:, None, None, None]
    return grad / grad_norm

  def inner(x):
    with x.graph.gradient_override_map({'Identity': op_name}):
      x = tf.identity(x)
    return x

  return inner


def compose(transforms):
  def inner(x):
    for transform in transforms:
      x = transform(x)
    return x
  return inner


def collapse_alpha_random(sd=0.5):
  def inner(t_image):
    rgb, a = t_image[..., :3], t_image[..., 3:4]
    rgb_shape = rgb.get_shape().as_list()
    rand_img = param.image_sample(rgb_shape, sd=sd)
    return a*rgb + (1-a)*rand_img
  return inner


def _rand_select(xs, seed=None):
  rand_n = tf.random_uniform((), 0, len(xs), "int32", seed=seed)
  return tf.constant(xs)[rand_n]


def _angle2rads(angle, units):
  angle = tf.cast(angle, "float32")
  if units.lower() == "degrees":
    angle = 3.14*angle/180.
  elif units.lower() in ["radians", "rads", "rad"]:
    angle = angle
  return angle


standard_transforms = [
    pad(12, mode='constant', constant_value=.5),
    jitter(8),
    random_scale([1 + (i-5)/50. for i in range(11)]),
    random_rotate(list(range(-10, 11)) + 5*[0]),
    jitter(4),
  ]
